/**
 *	@file stable_sort.h
 *	@brief	}[W\[g. watcompstable_sort̑pi.
 */
#ifndef STABLE_SORT_H
#define STABLE_SORT_H

#include <cstddef>
#include <algorithm>
#include <functional>
#include <iterator>
#ifndef assert
#include <cassert>
#endif


// }[W\[g̓. ͎gʖ.
namespace stable_sort_detail {

// }\[g. (T傫ƃX^bNɃRs[(tmp)邱ƂɂȂ̂ŁA ċAĂmerge_sort_body_OŎs)
template<typename BidiIte, class CMP>
void merge_sort_insertion_sort_(BidiIte first, BidiIte last, CMP cmp)
{
	typedef typename std::iterator_traits<BidiIte>::value_type T;
	BidiIte i = first;
	while (++i != last) {
		T		tmp( *i );
		BidiIte j = i;
		BidiIte k = i;
		do {
			--k;
			if (!cmp(tmp, *k))
				break;
			*j = *k;
			--j;
		} while (k != first);
		*j = tmp;
	}
}


template<typename BidiIte, class CMP, typename T>
void merge_sort_body_(BidiIte first, BidiIte last, CMP cmp, T* wk)
{
	std::size_t size = std::distance(first, last);
	if (size < 32) {	// ȉɂȂA}\[gŕבւ.
		if (size > 1)
			merge_sort_insertion_sort_(first, last, cmp);
		return;
	}

	// O㔼ɕʂă\[g.
	BidiIte 	mid( first);
	std::size_t halfsize = size / 2;
	assert(halfsize > 0);
	std::advance(mid, halfsize);

	merge_sort_body_(first, mid , cmp, wk); 	// O\[g.
	merge_sort_body_(mid  , last, cmp, wk); 	// 㔼\[g.

	// O㔼łɃ\[gς݂ł邱ƂOɕג.
	{
		BidiIte 	s( first );
		BidiIte 	d( first );
		T*			wk_e	 = wk + halfsize;
		T*			wp		 = wk;

		// [NɑORs[.
		while (wp != wk_e) {
			*wp++ = *s;
			++s;
		}

		// O㔼ׂ݂āÄ̗փf[^l߂Ă.
		wp = wk;
		do {
			if (cmp(*s, *wp)) {
				*d = *s;
				++s;
			} else {
				*d = *wp++;
			}
			++d;
		} while (s != last && wp != wk_e);

		if (wp != wk_e) {
			do {
				*d = *wp++;
				++d;
			} while (wp != wk_e);
		}
	  #if 0 // łɋl܂Ă͂.
		else {
			do {
				*d = *s;
				++d;
				++s;
			} while (s != last);
		}
	  #endif
	}
}


template<typename T>
struct merge_sort_work_mem_ {
	~merge_sort_work_mem_() { delete[] p_; }
	T*		p_;
};

}	// namespace stable_sort_detail



namespace std {

/// }[W\[g.
template<typename BidiIte, class CMP>
void stable_sort(BidiIte first, BidiIte last, CMP cmp)
{
	typedef typename std::iterator_traits<BidiIte>::value_type T;
	std::size_t 		size = std::distance(first, last);
	if (size < 48) {
		stable_sort_detail::merge_sort_insertion_sort_(first, last, cmp);
		return;
	}
	stable_sort_detail::merge_sort_work_mem_<T>  wk;
	wk.p_ = new T[ size / 2 ];
	stable_sort_detail::merge_sort_body_(first, last, cmp, wk.p_);
}


/// }[W\[g.
template<typename BidiIte> inline
void stable_sort(BidiIte first, BidiIte last) {
	stable_sort( first, last, std::less<typename std::iterator_traits<BidiIte>::value_type>() );
}

} // std

#endif
