/**
 *	@file	vecque_test.cpp
 *	@brief	vecque.h ̃eXg
 *	@author	Masashi Kitamura
 */

#include "unitst.h"
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <cassert>

#undef assert
#define assert(x)		UNITST_CHECK(x)
#include "../vecque.h"



using namespace std;



class VecQue_Test {
public:
	static void run() {
		test_array();
		test_pushpop();
		test_insert();
		test_sort();
	}

	static void test_array() {
		vecque<int> ary(5);

		UNITST_EQ(ary.size(), 5);
		UNITST_GE(ary.max_size(), 5);
		UNITST_GE(ary.capacity(), 5);
		UNITST_EQ(ary.empty()	, 0);

		UNITST_CHECK_ANY_THROW( (ary.at(5) = 5) );

		int n = 0;
		for (n = 0; n < 5; ++n)
			ary[n] = n;
		for (n = 0; n < 5; ++n)
			ary[n] = ary[n]+1;

		UNITST_EQ(ary.front(), 1);
		UNITST_EQ(ary.back(), 5);

		vecque<int>::iterator it;
		for (n = 1, it = ary.begin(); it != ary.end(); ++it, ++n) {
			UNITST_EQ(*it, n);
			++it;
			--it;
			it++;
			it--;
		}

		vecque<int>::reverse_iterator ri;
		for (n = 5, ri = ary.rbegin(); ri != ary.rend(); ++ri, --n)
			UNITST_EQ(*ri, n);

		vecque<int> ary2(ary);
		for (n = 0 ; n < 5; ++n)
			UNITST_EQ(ary2[n], ary[n]);
		UNITST_EQ(ary == ary2, true );
		UNITST_EQ(ary != ary2, false);
		UNITST_EQ(ary <  ary2, false);
		UNITST_EQ(ary <= ary2, true );
		UNITST_EQ(ary >  ary2, false);
		UNITST_EQ(ary >= ary2, true );

		vecque<int> ary3;
		ary3 = ary;
		for (n = 0 ; n < 5; ++n)
			UNITST_EQ(ary3[n], ary[n]);
		UNITST_EQ(ary3, ary);
		ary3[2] = 10;
		UNITST_EQ((ary == ary3), false);
		UNITST_EQ((ary != ary3), true );
		UNITST_EQ((ary <  ary3), true );
		UNITST_EQ((ary <= ary3), true );
		UNITST_EQ((ary >  ary3), false);
		UNITST_EQ((ary >= ary3), false);
		ary3[2] = 1;
		UNITST_EQ((ary == ary3), false);
		UNITST_EQ((ary != ary3), true );
		UNITST_EQ((ary <  ary3), false);
		UNITST_EQ((ary <= ary3), false);
		UNITST_EQ((ary >  ary3), true );
		UNITST_EQ((ary >= ary3), true );

		vecque<int> ary4 = ary3;

		ary2.swap(ary4);
		UNITST_EQ(ary2, ary3);
		UNITST_EQ(ary4, ary);

		swap(ary2, ary4);
		UNITST_EQ(ary2, ary);
		UNITST_EQ(ary4, ary3);

		vecque<int> ary5 = ary;
		ary5.reserve(10);
		for (n = 0 ; n < 5; ++n)
			UNITST_EQ(ary5[n], ary[n]);
		ary5.resize(10);

		for (n = 0; n < int(ary5.size()); ++n)
			ary5[n] = (n+1)*2;
		ary4.assign(ary5.begin(), ary5.begin()+ 5);
		// ary4 = ary5;
		for (n = 0 ; n < 5; ++n)
			UNITST_EQ(ary4[n], ary5[n]);

		ary4.assign(5U, (int)500);
		for (n = 0; n < 5; ++n)
			UNITST_EQ(ary4[n], 500);
	}


	static void test_pushpop() {
		vecque<int> v;
		int n;
		v.reserve(5);
		UNITST_EQ(v.size(), 0);
		UNITST_GE(v.max_size(), 5);
		UNITST_GE(v.capacity(), 5);
		UNITST_EQ(v.empty()   , 1);

		for (n = 0; n < 5; ++n) {
			v.push_back(n+1);
			UNITST_EQ( v[n] ,v.at(n) );
			UNITST_EQ(v.size(), unsigned(n+1));
			UNITST_GE(v.max_size(), 5);
			UNITST_GE(v.capacity(), 5);
			UNITST_EQ(v.empty()   , 0);
			UNITST_EQ(v.front(), 1);
			UNITST_EQ(v.back(), n+1);
		}
		//UNITST_CHECK_ANY_THROW( (v.at(5) = 5) );

		for (n = 5; --n >= 5;) {
			UNITST_EQ(v.back(), n+1);
			v.pop_back();
		}

		v.clear();
		UNITST_EQ(v.size(), 0);
		UNITST_EQ(v.empty()   , 1);

		v.resize(3);
		UNITST_EQ(v.size(), 3);
		UNITST_EQ(v.empty()   , 0);

		v.assign(3U, 999);
		for (n = 0; n < 3; ++n)
			UNITST_EQ(v[n], 999);

	}

	static void test_insert() {
		vecque<int> v;
		typedef vecque<int>::iterator vecque_ite;

	  #if 1
		v.insert( v.begin(), 1 );
	  #else
		v.resize(1);
		v.insert( v.begin(), 1 );
		v.erase(v.begin()+1);
	  #endif
		UNITST_EQ(v.size(), 1);
		UNITST_EQ(v.empty()   , 0);

		v.insert( v.end(), 5 );
		UNITST_EQ(v.size(), 2);
		UNITST_EQ(v.empty()   , 0);
		static const int m2[] = { 1, 5 };
		for (unsigned i = 0; i < 2; ++i)
			UNITST_EQ(v[i], m2[i]);

		v.insert( v.begin()+1, 3 );
		UNITST_EQ(v.size(), 3);
		UNITST_EQ(v.empty()   , 0);
		static const int m3[] = { 1, 3, 5 };
		for (unsigned i = 0; i < 3; ++i)
			UNITST_EQ(v[i], m3[i]);

		v.erase( v.begin()+1 );
		UNITST_EQ(v.size()	, 2);
		UNITST_EQ(v.empty() , 0);
		for (unsigned i = 0; i < 2; ++i)
			UNITST_EQ(v[i], m2[i]);

		static const int mi[] = {2,3,4};
	  #ifndef VECQUE_NO_MEMBER_TEMPLATES
		v.insert( v.begin()+1, mi, mi+3 );
	  #else
		v.insert( v.begin()+1, 1, mi[2] );
		v.insert( v.begin()+1, 1, mi[1] );
		v.insert( v.begin()+1, 1, mi[0] );
	  #endif
		UNITST_EQ(v.size()	, 5);
		UNITST_EQ(v.empty() , 0);
		static const int m5[] = { 1, 2, 3, 4, 5 };
		for (unsigned i = 0; i < 5; ++i)
			UNITST_EQ(v[i], m5[i]);

		v.erase( v.begin() );
		UNITST_EQ(v.size()	, 4);
		UNITST_EQ(v.empty() , 0);
		static const int m4a[] = { 2, 3, 4, 5 };
		//UNITST_EQ_MEM(&v[0], m4a, sizeof m4a);
		for (unsigned i = 0; i < 4; ++i)
			UNITST_EQ(v[i], m4a[i]);

		v.erase( v.begin()+1, v.end()-1 );
		UNITST_EQ(v.size()	, 2);
		UNITST_EQ(v.empty() , 0);
		static const int m2b[] = { 2, 5 };
		for (unsigned i = 0; i < 2; ++i)
			UNITST_EQ(v[i], m2b[i]);

	  #ifndef VECQUE_NO_MEMBER_TEMPLATES
		v.insert( v.end(), mi, mi+3 );
	  #else
		v.insert( v.end(), 1, mi[0] );
		v.insert( v.end(), 1, mi[1] );
		v.insert( v.end(), 1, mi[2] );
	  #endif
		UNITST_EQ(v.size()	, 5);
		UNITST_EQ(v.empty() , 0);
		v[0] = 1;
		static const int m5b[] = { 1, 5, 2, 3, 4, };
		for (unsigned i = 0; i < 5; ++i)
			UNITST_EQ(v[i], m5b[i]);

		vecque<int> v2;
		v2.reserve(10+2);
		v2.assign(v.begin(), v.end());
		UNITST_GE(v2.size()  , 5);
		UNITST_EQ(v2.empty() , 0);
		for (unsigned i = 0; i < 5; ++i)
			UNITST_EQ(v2[i], m5b[i]);

		v2.insert(v2.end(), v.begin(), v.end());
		static const int m10b[] = { 1, 5, 2, 3, 4, 1, 5, 2, 3, 4, };
		for (unsigned i = 0; i < 10; ++i)
			UNITST_EQ(v2[i], m10b[i]);

		v.erase( v.end()-2, v.end() );
		UNITST_EQ(v.size()	, 3);
		UNITST_EQ(v.empty() , 0);
		static const int m3b[] = { 1, 5, 2, };
		for (unsigned i = 0; i < 3; ++i)
			UNITST_EQ(v[i], m3b[i]);

		v.erase( v.begin(), v.end() );
		UNITST_EQ(v.size()	, 0);
		UNITST_EQ(v.empty() , 1);

		v.insert( v.begin(), 5U, 123 );
		UNITST_EQ(v.size()	, 5);
		UNITST_EQ(v.empty() , 0);
		static const int m5c[] = { 123, 123, 123, 123, 123 };
		for (unsigned i = 0; i < 5; ++i)
			UNITST_EQ(v[i], m5c[i]);

		v.insert( v.begin(), 1U, 1 );
		UNITST_EQ(v.size()	, 6);
		UNITST_EQ(v.empty() , 0);
		static const int m5d[] = { 1, 123, 123, 123, 123, 123 };
		for (unsigned i = 0; i < 6; ++i)
			UNITST_EQ(v[i], m5d[i]);

		//x v = v2;
		v.assign(v2.begin(), v2.begin() + v2.size());

#if 0
		UNITST_CHECK_ANY_THROW( v.insert( vecque_ite(), 1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		UNITST_CHECK_ANY_THROW( v.insert( vecque_ite()-1, 1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		UNITST_CHECK_ANY_THROW( v.insert( v.begin(), 1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);

		UNITST_CHECK_ANY_THROW( v.insert( v.begin()-1, m5c, m5c+3 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		UNITST_CHECK_ANY_THROW( v.insert( v.end()+1, m5c, m5c+3 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		UNITST_CHECK_ANY_THROW( v.insert( v.end(), m5c, m5c+1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);

		UNITST_PRINTF(("ȉ 4s̃G[bZ[Wΐ\n"));
		//UNITST_CHECK_ANY_THROW
		( v.erase( vecque_ite() ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		//UNITST_CHECK_ANY_THROW
		( v.erase( vecque_ite()-1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);

		//UNITST_CHECK_ANY_THROW
		( v.erase( v.begin()-1, v.end() ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);
		//UNITST_CHECK_ANY_THROW
		( v.erase( v.begin(), v.end()+1 ) );
		UNITST_EQ(v.size()	, 5);
		for (unsigned i = 0; i < 5; ++i) UNITST_EQ(v[i], m5b[i]);

		v.erase( v.end()-2, v.end() );
		UNITST_CHECK_ANY_THROW( v.insert( v.end(), m5b+5-2, m5b+5+1 ) );
		UNITST_EQ(v.size()	, 3);
		for (unsigned i = 0; i < 3; ++i) UNITST_EQ(v[i], m5b[i]);
#endif
	}

	struct rev_cmp {
		template<typename ITE>
		bool operator()(const ITE& l, const ITE& r) const {
			return r < l;
		}
	};

	static void test_sort() {
		vecque<int>	v;
		v.reserve(10);
		v.push_back(5);
		v.push_back(2);
		v.push_back(3);
		v.push_back(6);
		v.push_back(1);
		v.push_back(7);
		v.push_back(0);
		v.push_back(9);
		v.push_back(8);
		v.push_back(4);

		v.sort();
		for (int i = 0; i < 10; ++i)
			UNITST_EQ(v[i], i);

	  #if 1 //def VECQUE_NO_MEMBER_TEMPLATES
		stable_sort(v.begin(), v.end(), rev_cmp());
		for (int i = 0; i < 10; ++i)
			UNITST_EQ(v[i], 9-i);
	  #endif
	}

};


UNITST_TEST(vecque_test) {
	VecQue_Test::run();
}
